import { useReducer } from 'react';

interface State<T> {
  set: Set<T>;
}
interface Action<T> {
  type: 'add' | 'delete' | 'clear';
  value?: T;
}

function createReducer<T>() {
  return function (state: State<T>, action: Action<T>) {
    switch (action.type) {
      case 'add': {
        state.set.add(action.value!);
        break;
      }
      case 'delete': {
        state.set.delete(action.value!);
        break;
      }
      case 'clear': {
        state.set.clear();
        break;
      }
      default:
        break;
    }
    return {
      set: state.set,
    };
  };
}

export function useSet<T>(arr: T[] = []) {
  const [state, dispatch] = useReducer(createReducer<T>(), {
    set: new Set<T>(arr),
  });

  return {
    add: (value: T) => {
      dispatch({ type: 'add', value });
      return value;
    },
    delete: (value: T) => {
      dispatch({ type: 'delete', value });
      return value;
    },
    clear: () => dispatch({ type: 'clear' }),
    forEach: (callback: (value: T, key: T, set: Set<T>) => void) =>
      state.set.forEach(callback),
    map: (callback: <K>(value: T, index: number, arr: T[]) => K) =>
      [...state.set].map(callback),
    has: (value: T) => state.set.has(value),
    entries: () => state.set.entries(),
    values: () => state.set.values(),
    keys: () => state.set.keys(),
    size: state.set.size,
  };
}
